#!/bin/bash
########################################################################
# build_pythia8.sh
#
# Build pythia8 and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_11_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}
thisdir=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" || ssi_die "unable to setup UPS"
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. ${thisdir}/version_info.sh || \
  ssi_die "unable to load version info from ${thisdir}/version_info"
####################################

##################
# Other useful variables. *C*
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,c7} \
  --pqual={p372,p383b,p392} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

# Needs to be early so we can give a useful error message rather than
# having the UPS setup attempt fail.
if [[ "${pqual}" =~ ^p([[:digit:]]+) ]] && \
  { (( ${BASH_REMATCH[1]:0:1} == 2 )) || (( ${BASH_REMATCH[1]:1} < 83 )); } && \
  { { [ "${cqual:0:1}" = e ] && (( ${cqual#e} > 19 )); } || \
    { [ "${cqual:0:1}" = c ] && (( ${cqual#c} > 7 )); } || \
    ! [[ "${cqual:0:1}" == [ce] ]]; }; then
  case ${cqual:0:1} in
    e) emsg="GCC > 8.3";;
    c) emsg="LLVM/Clang > 7.0.0";;
    *) emsg="non-GCC / LLVM compilers"
  esac
  ssi_die "Python < 3.8.3 not supported with ${emsg}"
fi

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS IN_SOURCE || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
####################################

# Print active products.
echo
ups active
echo

# Early so we can use ${sitearch} in is_built().
case ${pqual} in
  p372|p383b|p392) ${trace} python_vars || exit
    python_plugin="${pkgdir}/lib/${package}.so"
    absolute_path -- "${pkgdir}/lib/${PYTHON_LIBDIR}/site-packages" sitearch;;
  *) ssi_die -I "accepted unknown pqual=${pqual}"
esac

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  fix-lcg-detection.patch \
  fix-makefile-sed.patch \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

cxxflg=()
case ${cqual} in
  e19|e20) cc=gcc; cxx=g++; cxxstd=17;;
  c7) cc=clang; cxx=clang++; cxxstd=17;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac
cxxflg+=( \
  -std=c++${cxxstd} -fPIC -pedantic \
  -W -Wall -Wshadow -Wno-deprecated-declarations \
  )
case ${extraqual} in
  debug) cxxflg+=(-g -O0);;
  prof) cxxflg+=(-g -O3 -DNDEBUG -fno-omit-frame-pointer);;
  *) ssi_die -I "accepted unknown btype=${btype}"
esac

[ "${pqual}" = "p2715a" ] && cxxflg+=(-Wno-register)

##################
# Configure parallelism. *C*
ncores=$(ncores)
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}/${pkgtarname}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

# Notes:
#
# * --cxx-common overrides the default options (including -fPIC) so
#   we need to be very specific and complete.

${trace} "${sourcedir}/${pkgtarname}/configure" --prefix="${pkgdir}" \
  --cxx=${cxx} \
  --cxx-common="${cxxflg[*]}" \
  --with-gzip \
  --with-python-include="${PYTHON_INCLUDE}" \
  --with-lhapdf6="${LHAPDF_FQ_DIR}" \
  --with-hepmc3-lib="${HEPMC3_LIB}" \
  || ssi_die "configure failed"
##################

##################
# Build. *C*

${trace} make -j ${ncores} ${BUILD_DEBUG:+VERBOSE=1} || \
  ssi_die "build failed"
##################

##################
# Test. *C*
##################

##################
# Install. *C*

${trace} make -j ${ncores} install || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*

# Tidy source for installation as debug source.
${trace} make clean; ${trace} find . -name '*~' | xargs rm -f

# Move cleaned source to installation point.
ssi_info "installing debug source"
${trace} cd "${pkgtop}" && \
  ${trace} rm -rf source && \
  ${trace} mv "${blddir}" source || \
    ssi_die "unable to move source for packaging"

# Relocate the Python plugin to its final resting place.
if [[ -f "${python_plugin}" ]]; then
  ${trace} mkdir -p "${sitearch}" && \
  ${trace} mv -v "${python_plugin}" "${sitearch}" || \
    ssi_die "unable to move Python plugin ${python_plugin##*/} into ${sitearch}/"
fi

ssi_info "post-processing for relocatability"
${trace} fix_config_script \
  =PYTHIA8_FQ_DIR \
  =PYTHIA8_INC \
  =PYTHIA8_LIB \
  =LHAPDF_FQ_DIR \
  =LHAPDF_LIB \
  =LHAPDF_INC \
  =HEPMC3_FQ_DIR \
  =HEPMC3_LIB \
  =HEPMC3_INC \
  =PYTHONHOME \
  =PYTHON_INCLUDE \
  =PYTHON_LIB \
  =PYTHON_LIBDIR \
  "${pkgdir}/bin/pythia8-config" \
  "${pkgdir}/share/Pythia8/examples/Makefile.inc" || \
  ssi_die "post-processing failed"
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
